from scrapeops_scrapy.normalizer.domains import DomainNormalizer
from scrapeops_scrapy.normalizer.proxies import ProxyNormalizer
from scrapeops_scrapy.normalizer.proxy_port_normalizer import ProxyPortStringNormalizer


class BaseRequestResponse(object):
    """
        Normalised request/response data structure.
    """

    def __init__(self):
        self.signal_type = None
        self.request = None
        self.raw_url = None
        self.raw_proxy_port = None
        self.raw_domain = None
        self.raw_headers = None
 
        ## Proxy Checks
        self._active_proxy = None
        self._real_url = None
        self._ip_proxy_list = False
        self._named_proxy = False

        ## Proxy Port
        self._proxy_port_name = None
        self._complete_proxy_port_string = None
        self._proxy_setup_key = None

        self._proxy_port_scheme = ''
        self._proxy_port_username = ''
        self._proxy_port_password = ''
        self._proxy_port_host = ''
        self._proxy_port_port = ''
        self._proxy_port_headers = {}

        self._normalized_proxy_port_username = None
        self._normalized_proxy_port_password = None
        self._normalized_proxy_port_host = None
        self._normalized_proxy_port_port = None
        self._normalized_proxy_port_header_string = None


        ## Proxy API
        self._proxy_api = False
        self._proxy_api_name = None

        ## Validation
        self._validation_test = None
        self._geo = None
        self._custom_tag = None 

        ## Final
        self._domain = None
        self._page_type = None
        self._proxy_type = None
        self._proxy_name = None
        self._proxy_setup = None
        

    """
        Getters
    """

    def get_proxy_name(self):
        return self._proxy_name or 'unknown'
    
    def get_proxy_setup(self):
        return self._proxy_setup or 'unknown'

    def get_domain(self):
        return self._domain or 'unknown'
    
    def get_page_type(self):
        return self._page_type or 'unknown'

    def get_proxy_api_name(self):
        return self._proxy_api_name

    def get_proxy_port_name(self):
        return self._proxy_port_name

    def get_raw_proxy(self):
        return self.raw_proxy_port

    def get_real_url(self):
        return self._real_url or 'unknown'

    def get_validation_test(self):
        return self._validation_test or 'pass'

    def get_geo(self):
        return self._geo or 'none'

    def get_custom_tag(self):
        return self._custom_tag or 'none'

    def get_proxy_port_username(self):
        return self._proxy_port_username

    def get_proxy_port_password(self):
        return self._proxy_port_password

    def get_proxy_port_host(self):
        return self._proxy_port_host

    def get_proxy_port_port(self):
        return self._proxy_port_port
    
    def get_proxy_port_headers(self):
        if self._proxy_port_headers == {}:
            self._proxy_port_headers = ProxyNormalizer.convert_headers(self.raw_headers)
        return self._proxy_port_headers

    def get_complete_proxy_string(self):
        if self._complete_proxy_port_string is None:
            self._complete_proxy_port_string = "{}://{}:{}@{}:{}".format(self._proxy_port_scheme, self._proxy_port_username, self._proxy_port_password,
                                                                            self._proxy_port_host, self._proxy_port_port)
        return self._complete_proxy_port_string

    def get_normalized_proxy_port_username(self):
        if self._normalized_proxy_port_username is None:
            return self._proxy_port_username
        return self._normalized_proxy_port_username

    def get_normalized_proxy_port_password(self):
        if self._normalized_proxy_port_password is None:
            return self._proxy_port_password
        return self._normalized_proxy_port_password

    def get_normalized_proxy_port_host(self):
        if self._normalized_proxy_port_host is None:
            return self._proxy_port_host
        return self._normalized_proxy_port_host

    def get_normalized_proxy_port_port(self):
        if self._normalized_proxy_port_port is None:
            return self._proxy_port_port
        return self._normalized_proxy_port_port

    def get_normalized_proxy_port_header_string(self):
        if self._normalized_proxy_port_header_string is not None:
            return f' -H {self._normalized_proxy_port_header_string}'
        return ''


    """
        SETTERS
    """

    def set_normalized_proxy_port_username(self, username):
        self._normalized_proxy_port_username = username

    def set_normalized_proxy_port_password(self, password):
        self._normalized_proxy_port_password = password

    def set_normalized_proxy_port_host(self, host):
        self._normalized_proxy_port_host = host

    def set_normalized_proxy_port_port(self, port):
        self._normalized_proxy_port_port = port

    def update_normalized_proxy_port_header_string(self, header_string):
        if self._normalized_proxy_port_header_string is None:
            self._normalized_proxy_port_header_string = header_string
        else:
            self._normalized_proxy_port_header_string = f'{self._normalized_proxy_port_header_string} {header_string}' 

        
    """
        Proxy Type Methods
    """
    
    def active_proxy(self):
        return True if self._active_proxy else False
    
    def active_proxy_port(self):
        return True if self._active_porxy_port else False

    def active_proxy_api(self):
        return self._proxy_api

    def active_named_proxy(self):
        return self._named_proxy

    




class RequestResponse(BaseRequestResponse):

    def __init__(self, signal_type=None, request=None, response=None):
        BaseRequestResponse.__init__(self)
        self.signal_type = signal_type
        if request is not None or response is not None:
            self.request = response.request if request is None else request
            self.raw_url = request.url if response is None else response.url
            self.raw_proxy_port = self.request.meta.get('proxy') 
            self.raw_domain = DomainNormalizer.get_domain(self.raw_url)
            self._active_proxy = self._active_porxy_port = False if self.raw_proxy_port is None else True
            self.raw_headers = self.request.headers

    """
        Domain Normalization
    """

    def check_domain(self, domain_obj):
        domain_details = domain_obj.get(self._domain)
        if domain_details is not None:
            self._page_type = DomainNormalizer.get_page_type(self._real_url, domain_data=domain_details)
            return False
        return True


    def update_page_type(self, domain_details):
        if domain_details is not None:
            self._page_type = DomainNormalizer.get_page_type(self._real_url, domain_data=domain_details)


    def fallback_domain_data(self):
        if self._domain is None:
            self._domain = DomainNormalizer.get_domain(self.raw_url)
        self._page_type = 'none'


    """
        Proxy Port Normalization
    """

    def check_proxy_port_type(self, proxy_ports):
        if ProxyNormalizer.check_ip_address(self.raw_proxy_port):
            self._proxy_type = 'proxy_ip_list'
            self._real_url = self.raw_url
            self._domain = self.raw_domain
            self._proxy_name = 'unknown_ip'
            self._proxy_setup = 'ip_address'
            return False, False
        
        self._named_proxy, self._proxy_port_host, self._proxy_port_name = ProxyNormalizer.check_named_proxy(self.raw_proxy_port) 

        if self._named_proxy:
            self._proxy_type = 'named_proxy_port'
            self._real_url = self.raw_url
            self._domain = self.raw_domain
            self.get_proxy_port_details()
            
            proxy_details = proxy_ports.get(self._proxy_port_name)
            
            if proxy_details is not None:

                if proxy_details.get(self._complete_proxy_port_string) is not None:
                    self._proxy_setup = proxy_details.get(self._complete_proxy_port_string) 
                elif proxy_details.get(self._complete_proxy_port_string) is None and proxy_details.get('known', False):
                    ProxyPortStringNormalizer.run_proxy_string_normalization(self, proxy_ports[self._proxy_port_name].get('normalization_actions'))
                    self.create_normalized_proxy_port_string()
                    self._proxy_setup = proxy_details.get(self._normalized_proxy_port_string) 

                if self._proxy_setup is None:
                    self._proxy_setup = proxy_details.get('fallback') 
                    if proxy_details.get('count') > proxy_details.get('max_count') or proxy_details.get('known') is False:
                        return True, False
                    ## Get details
                    return True, True
                return True, False

            ## get proxy details
            return True, True
    

    def get_proxy_port_details(self):
        self._proxy_name = self._proxy_port_name
        self._proxy_port_port = ProxyNormalizer.get_proxy_port(self.raw_proxy_port)
        self._proxy_port_scheme = ProxyNormalizer.get_proxy_scheme(self.raw_proxy_port)
        if self.raw_headers.get('Proxy-Authorization') is not None:
                auth_string = self.raw_headers.get('Proxy-Authorization').decode('utf-8') 
                self._proxy_port_username, self._proxy_port_password = ProxyNormalizer.decode_basic_auth(auth_string)
        self._complete_proxy_port_string = "{}://{}:{}@{}:{}".format(self._proxy_port_scheme, self._proxy_port_username, self._proxy_port_password,
                                                                            self._proxy_port_host, self._proxy_port_port)

    def create_normalized_proxy_port_string(self):
        username = self.get_normalized_proxy_port_username()
        password = self.get_normalized_proxy_port_password()
        host = self.get_normalized_proxy_port_host()
        port = self.get_normalized_proxy_port_port()
        header_string = self.get_normalized_proxy_port_header_string()
        self._normalized_proxy_port_string = "{}://{}:{}@{}:{}".format(self._proxy_port_scheme, username, password, host, port)
        if header_string != '':
            self._normalized_proxy_port_string = self._normalized_proxy_port_string + header_string

    def proxy_port_setup(self, proxy_details):
        proxy_setup = proxy_details.get('proxy_setup')
        if proxy_setup is None:
            return 'none'
        proxy_string = 'port'
        ## Generate settings string
        return proxy_string

    def update_proxy_port(self, proxy_name, proxy_setup_value):
        self._active_proxy = True
        self._proxy_api = False
        self._proxy_type = 'named_proxy_port'
        self._proxy_name = proxy_name
        self._proxy_setup = proxy_setup_value



    """
        Proxy API Normalization
    """

    def check_proxy_api(self, proxy_apis):
        proxy_details = proxy_apis.get(self.raw_domain)
        if proxy_details is not None:
            if proxy_details.get('proxy_setup') is None:
                self._proxy_api_name = proxy_details.get('proxy_name')
                return True, True
            self.update_proxy_api(proxy_details)
            return True, False
        return False, False


    def update_proxy_api(self, proxy_details):
        self._real_url = DomainNormalizer.get_url_proxy_api(url=self.raw_url, proxy_settings=proxy_details)
        self._domain = DomainNormalizer.get_domain(self._real_url)
        self._active_proxy = True
        self._proxy_api = True
        self._proxy_type = 'proxy_api'
        self._proxy_name = self._proxy_api_name = proxy_details.get('proxy_name')
        self._proxy_setup = self.proxy_api_setup(proxy_details) ## into new file


    def proxy_api_setup(self, proxy_details):
        proxy_string = 'api'
        proxy_setup = proxy_details.get('proxy_setup')
        if proxy_setup is None:
            return proxy_string
        query_params = DomainNormalizer.parse_url(self.raw_url)
        for key, value in query_params.items():
            key_mapping = proxy_setup.get(key)
            if key_mapping is not None:
                if key_mapping.startswith('**'):
                    proxy_string = f'{proxy_string}_{key_mapping[2:]}'
                elif key_mapping.startswith('--'):
                    proxy_string = f'{proxy_string}_{key_mapping[2:]}={value.lower()}'
                elif key_mapping.startswith('^^'):
                    proxy_string = f'{proxy_string}_{key_mapping[2:]}=false'
                else:
                    proxy_string = f'{proxy_string}_{key_mapping}=true'
        return proxy_string

    
    
    """
        Fallback Proxy Details
    """

    def update_no_proxy(self):
        self._proxy_type = self._proxy_name = 'no_proxy'
        self._proxy_setup = 'none'
        self._real_url = self.raw_url
        self._domain = self.raw_domain
    
    def fallback_proxy_details(self, proxy_type=None, proxy_apis=None):
        if proxy_type == 'proxy_api':
            proxy_details = proxy_apis.get(self.raw_domain)
            if proxy_details is not None:
                self.update_proxy_api(proxy_details)
            else:
                self._proxy_name = 'unknown_proxy_api' if self._proxy_api_name is None else self._proxy_api_name
                self._proxy_setup = 'fallback' if self._proxy_setup is None else self._proxy_setup
        else:
            self._proxy_name = 'unknown_proxy_port' if self._proxy_name is None else self._proxy_name
            self._proxy_setup = 'fallback' if self._proxy_setup is None else self._proxy_setup


    """
        Fallback Proxy + Domain Details
    """
    
    def fallback_domain_proxy_details(self, reason='fallback'):
        """
            Fallback -> if issue with domain/proxy normalising
        """
        self._domain = DomainNormalizer.get_domain(self.raw_url)
        self._page_type = 'none'
        self._proxy_name = reason
        self._proxy_setup = 'none'

    
    """
        Response Validation Tests
    """

    def failed_validation_test(self, test):
        if self._validation_test is None:
            self._validation_test = test.get('validation_msg', 'failed')
        else:
            self._validation_test = f'{self._validation_test}&&{test.get("validation_msg", "failed")}' 
        if test.get('validation_test_id', -1) != -1:
          self._validation_test = f'{self._validation_test}_{test.get("validation_test_id")}'   
















        
        


        

        






import re
import socket

from base64 import b64decode
from urllib.parse import unquote, urlparse

from scrapeops_scrapy.normalizer.domains import DomainNormalizer
from scrapeops_scrapy.exceptions import DecodeError


class ProxyNormalizer(object):

    def __init__(self):
        pass
    
    @staticmethod
    def check_named_proxy(proxy_string):
        try: 
            proxy_address = DomainNormalizer.get_full_domain(proxy_string)
            proxy_domain = DomainNormalizer.get_domain(proxy_string)
            return True, proxy_address, proxy_domain
        except Exception:
            return False, 'ip_list_proxy'

    @staticmethod
    def remove_brackets(string):
        characters_to_remove = ['[',']']
        new_string = string
        for character in characters_to_remove:
            new_string = new_string.replace(character, "")
        return new_string
    
    
    @staticmethod
    def check_ip_address(proxy_string):
        s = ProxyNormalizer.remove_brackets(proxy_string)
        ipv6_split_string = re.split('://|@|/', s)
        for el in ipv6_split_string:
            if ProxyNormalizer.is_valid_ipv6_address(el): return True
            
        ipv4_split_string = re.split('://|:|@', proxy_string)
        for el in ipv4_split_string:
            if ProxyNormalizer.is_valid_ipv4_address(el): return True
        return False


    @staticmethod
    def is_valid_ipv4_address(address):
        try:
            socket.inet_pton(socket.AF_INET, address)
        except AttributeError:  # no inet_pton here, sorry
            try:
                socket.inet_aton(address)
            except socket.error:
                return False
            return address.count('.') == 3
        except socket.error:  # not a valid address
            return False

        return True

    @staticmethod
    def is_valid_ipv6_address(address):
        try:
            socket.inet_pton(socket.AF_INET6, address)
        except socket.error:  # not a valid address
            return False
        return True

    @staticmethod
    def get_proxy_port(proxy_string):
        try:
            return urlparse(proxy_string).port
        except Exception:
            return '80'  
    
    @staticmethod
    def get_proxy_host(proxy_string):
        try: 
            return DomainNormalizer.get_full_domain(proxy_string)
        except Exception:
            return 'ip_list_proxy'

    @staticmethod
    def get_proxy_scheme(proxy_string):
        try:
            return urlparse(proxy_string).scheme
        except Exception:
            return ''

    @staticmethod
    def unknown_proxy_scheme(proxy_string):
        if ProxyNormalizer.get_proxy_scheme(proxy_string) == '':
            return True
        return False

    @staticmethod
    def convert_bytes_to_string(inputValue):
        if isinstance(inputValue, (str, int)):
            return inputValue 
        if isinstance(inputValue, (bytes, bytearray)):
            return inputValue.decode('utf-8') 
        if isinstance(inputValue, list):
            tempList = []
            for el in inputValue:
                if isinstance(el, (bytes, bytearray)):
                   tempList.append(el.decode('utf-8')) 
                elif isinstance(el, list):
                    tempList.append([''])
                elif isinstance(el, dict):
                    tempList.append({'': ''})
                else:
                    tempList.append(el)
            return tempList
        return inputValue

    @staticmethod
    def convert_headers(raw_headers):
        header_dict = {}
        try:
            for key, value in raw_headers.items():
                k = ProxyNormalizer.convert_bytes_to_string(key)
                v = ProxyNormalizer.convert_bytes_to_string(value)
                header_dict[k] = v 
            return header_dict
        except Exception:
            return header_dict

    @staticmethod
    def decode_basic_auth(auth_string):
        """Decode an encrypted HTTP basic authentication string. Returns a tuple of
        the form (username, password), and raises a DecodeError exception if
        nothing could be decoded.
        """
        split = auth_string.strip().split(' ')

        # If split is only one element, try to decode the username and password
        # directly.
        if len(split) == 1:
            try:
                username, password = b64decode(split[0]).decode().split(':', 1)
            except Exception:
                raise DecodeError

        # If there are only two elements, check the first and ensure it says
        # 'basic' so that we know we're about to decode the right thing. If not,
        # bail out.
        elif len(split) == 2:
            if split[0].strip().lower() == 'basic':
                try:
                    username, password = b64decode(split[1]).decode().split(':', 1)
                except Exception:
                    raise DecodeError
            else:
                raise DecodeError

        # If there are more than 2 elements, something crazy must be happening.
        # Bail.
        else:
            raise DecodeError

        return unquote(username), unquote(password)

    @staticmethod
    def create_dict_if_none_exists(dict, key):
        if dict.get(key) is None:
            dict[key] = {}

    @staticmethod
    def update_proxy_details(proxy_dict, request_response_object, data, valid=False):
        proxy_name = request_response_object.get_proxy_port_name()
        if proxy_dict.get(proxy_name) is None:
            proxy_dict[proxy_name] = {}

        ## Update counter
        proxy_port_details = data.get('proxy_port_details')
        count = proxy_dict[proxy_name].get('count', 0)
        proxy_dict[proxy_name]['count'] = count + 1
        proxy_dict[proxy_name]['max_count'] = proxy_port_details.get('max_count', 3)

        if valid:
            proxy_dict[proxy_name]['normalization_actions'] = data.get('normalization_actions')
            proxy_dict[proxy_name]['fallback'] = data.get('fallback', 'port')

            
            proxy_setup_key = proxy_port_details.get('proxy_setup_key')
            proxy_setup_value = proxy_port_details.get('proxy_setup_value')
            if proxy_setup_value is None: 
                proxy_setup_value = data.get('fallback', 'port_type=unknown')
            proxy_dict[proxy_name][proxy_setup_key] = proxy_setup_value
            proxy_dict[proxy_name]['known'] = proxy_port_details.get('proxy_known_domain', False)
            request_response_object.update_proxy_port(proxy_name, proxy_setup_value)

        
        else:
            proxy_dict[proxy_name]['fallback'] = 'port'
            request_response_object.fallback_proxy_details(proxy_type='proxy_port')

        

    

    


    
 



   


    







         
    



import time
import sys
import sys
import scrapy
import scrapeops_scrapy
import platform

from scrapeops_scrapy.utils.error_handling import exception_handler

def current_time():
    t = time.time()
    return int(round(t, 0))

@exception_handler
def get_args():
    arg_dict = {'raw_string': '', 'args': [], 'options': []}
    if sys.argv[0] == 'crawl' or sys.argv[0]  == 'runspider': 
        args = sys.argv[2:]
    else:
        args = sys.argv[1:]
    for index, arg in enumerate(args):
        arg_dict['raw_string'] += append_raw_string(arg)
        if arg.startswith('--'):
            arg_dict['options'].append(arg)
        if arg.startswith('-a'):
            try:                   
                if args[index + 1].startswith('-') is False and args[index + 1].startswith('--') is False: arg_dict['args'].append(args[index + 1])  
            except Exception:
                arg_dict['args'].append(arg)
    return arg_dict


def scrapeops_middleware_installed(spider_settings):
    downloader_middlerwares = spider_settings.get('DOWNLOADER_MIDDLEWARES', {})
    if downloader_middlerwares.get('scrapeops_scrapy.middleware.stats.ScrapeOpsStats') is not None:
        return True
    if downloader_middlerwares.get('scrapeops_scrapy.middleware.retry.RetryMiddleware') is not None:
        return True
    return False

@exception_handler
def get_python_version():
    verions_string = sys.version
    split_string = verions_string.split(' ')
    return split_string[0]

@exception_handler
def get_scrapy_version():
    return scrapy.__version__

@exception_handler
def get_scrapeops_version():
    return scrapeops_scrapy.__version__

@exception_handler
def get_system_version():
    return platform.platform()

def append_raw_string(arg):
    if ' ' in arg:
         return '"{}"  '.format(arg)
    return "{}  ".format(arg)

def merge_dicts(x, y):
    z = x.copy()   
    z.update(y)
    return z









class FailedUrlsHandler(object):

    FAILED_URL_LOGGER_ENABLED = True
    LOG_MISSED_URLS = False
    MAX_LOGGED_URLS = 100

    def __init__(self):
        self.failed_urls_count = 0
        self.failed_urls_list = []
        self.errback_free = True

    def log_failure(self, failure):
        if FailedUrlsHandler.FAILED_URL_LOGGER_ENABLED:
            self.failed_urls_count += 1
            if FailedUrlsHandler.LOG_MISSED_URLS and len(self.failed_urls_list) < FailedUrlsHandler.MAX_LOGGED_URLS:
                request = failure.request
                self.failed_urls_list.append(request.url)

    def get_url_count(self):
        return self.failed_urls_count

    def get_url_list(self):
        return self.failed_urls_list

    def disable_errback(self):
        self.errback_free = False

    def enabled(self):
        return self.errback_free





from scrapeops_scrapy.exceptions import ScrapeOpsAPIResponseError
from scrapeops_scrapy.utils.utils import merge_dicts
from scrapeops_scrapy.normalizer.proxies import ProxyNormalizer
import requests
import time

    
class SOPSRequest(object):

    TIMEOUT = 30
    RETRY_LIMIT = 3
    API_KEY = None
    JOB_GROUP_ID = None
    SCRAPEOPS_ENDPOINT = 'https://api.scrapeops.io/'
    SCRAPEOPS_API_VERSION = 'api/v1/'
    SCRAPEOPS_LOGGING_DATA = None

    def __init__(self):
        self.data = None
        self.valid = None
        self.action = None
        self.error = None
        
    
    def setup_request(self, body=None):
        url = SOPSRequest.SCRAPEOPS_ENDPOINT + SOPSRequest.SCRAPEOPS_API_VERSION + f'setup/'
        data, error = SOPSRequest.post(url, body=body)
        data, self.valid, self.action, self.error = SOPSRequest.setup_stats_validation(data, error)
        return data, self


    def stats_request(self, body=None, log_body=None, files=None):
        post_body = merge_dicts(SOPSRequest.SCRAPEOPS_LOGGING_DATA, body)
        if files is not None:
           url = SOPSRequest.SCRAPEOPS_ENDPOINT + SOPSRequest.SCRAPEOPS_API_VERSION + f'logs/?log_type=scrapy' 
           _, _ = SOPSRequest.post_file(url, body=log_body, files=files)
        url = SOPSRequest.SCRAPEOPS_ENDPOINT + SOPSRequest.SCRAPEOPS_API_VERSION + f'stats/'
        data, error = SOPSRequest.post(url, body=post_body)
        data, self.valid, self.action, self.error = SOPSRequest.setup_stats_validation(data, error)
        return data, self


    def error_report_request(self, error_type=None, body=None, files=None):
        post_body = merge_dicts(SOPSRequest.SCRAPEOPS_LOGGING_DATA, body)
        if files is None:
            url = SOPSRequest.SCRAPEOPS_ENDPOINT + SOPSRequest.SCRAPEOPS_API_VERSION + f'errors/?error_type={error_type}'
            data, error = SOPSRequest.post(url, body=post_body, files=files)
        else:
            url = SOPSRequest.SCRAPEOPS_ENDPOINT + SOPSRequest.SCRAPEOPS_API_VERSION + f'errors/logs/?error_type={error_type}'
            data, error = SOPSRequest.post_file(url, body=post_body, files=files)
        data, self.valid, self.action, self.error = SOPSRequest.error_report_validation(data, error)
        return data, self
    

    def proxy_normalisation_request(self, request_response_object):
        proxy_name = request_response_object.get_proxy_port_name() 
        proxy_string = request_response_object.get_raw_proxy()
        post_body = merge_dicts(SOPSRequest.SCRAPEOPS_LOGGING_DATA, {'proxy_string': proxy_string})
        url = SOPSRequest.SCRAPEOPS_ENDPOINT + SOPSRequest.SCRAPEOPS_API_VERSION + f'normalizer/proxy/?proxy_name={proxy_name}'
        data, error = SOPSRequest.post(url, body=post_body)
        data, self.valid, self.action, self.error = SOPSRequest.normaliser_validation(data, error, request_type='proxy')
        return data, self


    def proxy_api_normalisation_request(self, request_response_object):
        proxy_name = request_response_object.get_proxy_api_name()
        url = SOPSRequest.SCRAPEOPS_ENDPOINT + SOPSRequest.SCRAPEOPS_API_VERSION + f'normalizer/proxy_api/?proxy_name={proxy_name}'
        data, error = SOPSRequest.post(url, body=SOPSRequest.SCRAPEOPS_LOGGING_DATA)
        data, self.valid, self.action, self.error = SOPSRequest.normaliser_validation(data, error, request_type='proxy_api')
        return data, self

    def proxy_port_normalisation_request(self, request_response_object, test_data=None):
        post_body = merge_dicts(SOPSRequest.SCRAPEOPS_LOGGING_DATA, {
            'proxy_string': request_response_object.get_complete_proxy_string(), 
            'proxy_headers': request_response_object.get_proxy_port_headers(), 
            'domain': request_response_object.get_proxy_port_name()})
        url = SOPSRequest.SCRAPEOPS_ENDPOINT + SOPSRequest.SCRAPEOPS_API_VERSION + f'normalizer/proxy_port/?job_id={SOPSRequest.JOB_GROUP_ID}'
        if test_data is not None:
            post_body['test_data'] = test_data
        data, error = SOPSRequest.post(url, body=post_body)
        data, self.valid, self.action, self.error = SOPSRequest.normaliser_validation(data, error, request_type='proxy_port')
        return data, self


    def domain_normalisation_request(self, request_response_object):
        domain = request_response_object.get_domain()
        real_url = request_response_object.get_real_url()
        post_body = merge_dicts(SOPSRequest.SCRAPEOPS_LOGGING_DATA, {'url': real_url})
        url = SOPSRequest.SCRAPEOPS_ENDPOINT + SOPSRequest.SCRAPEOPS_API_VERSION + f'normalizer/domain/?domain={domain}'
        data, error = SOPSRequest.post(url, body=post_body)
        data, self.valid, self.action, self.error = SOPSRequest.normaliser_validation(data, error, request_type='domain')
        return data, self

    def proxy_alert_request(self, request_response_object, job_group_id, error_response, alerts_sent):
        data = error_response
        data['domain'] = request_response_object.get_domain()
        data['proxy_provider'] = request_response_object.get_proxy_name()
        data['alerts_sent'] = alerts_sent
        post_body = merge_dicts(SOPSRequest.SCRAPEOPS_LOGGING_DATA, data)
        url = SOPSRequest.SCRAPEOPS_ENDPOINT + SOPSRequest.SCRAPEOPS_API_VERSION + f'alerts/proxy/?job_group_id={job_group_id}'
        data, error = SOPSRequest.post(url, body=post_body)
        data, self.valid, self.error = SOPSRequest.generic_validation(data, error)
        return data, self

    def proxy_test_request(self, url, request_response_object):
        data, _ = SOPSRequest.get(url, proxy=request_response_object.get_complete_proxy_string())
        return data

    @staticmethod
    def generic_validation(data, error):
        if data is None:
            return data, False, str(error)
        elif data.get('api_key') == 'invalid':
            return data, False, 'invalid_api_key'
        elif data.get('job_id') == 'invalid':
            return data, False, 'invalid_job'
        return data, True, None


    @staticmethod
    def setup_stats_validation(data, error):
        if data is None:
            return data, False, 'retry', str(error)
        elif data.get('api_key') == 'invalid':
            return data, False, 'close', 'invalid_api_key'
        elif data.get('job_valid') is not True and data.get('job_id') is None:
            return data, False, 'retry', 'invalid_job'
        return data, True, 'valid', None


    @staticmethod
    def normaliser_validation(data, error, request_type=None):
        if data is None:
            return data, False, 'fallback', str(error)
        elif data.get('api_key') == 'invalid':
            return data, False, 'close', 'invalid_api_key'
        
        ## proxy port 
        elif request_type=='proxy_port' and data.get('proxy_port_details') is None:
            return data, False, 'fallback', 'no_proxy_port_details'

        ## proxy api
        elif request_type=='proxy_api' and data.get('proxy_parsing_data') is None:
            return data, False, 'fallback', 'no_proxy_parsing_data'
        elif request_type=='proxy_api' and data.get('proxy_parsing_data') is not None:
            proxy_parsing_data = data.get('proxy_parsing_data')
            if proxy_parsing_data.get('known_proxy') is False:
                return data, False, 'fallback', 'unknown_proxy'
        
        ## domain specific
        elif request_type=='domain' and data.get('domain_parsing_data') is None:
            return data, False, 'fallback', 'no_domain_parsing_data'
        return data, True, 'valid', None

    
    @staticmethod
    def error_report_validation(data, error):
        if data is None:
            return data, False, 'retry', str(error)
        elif data.get('error_logged') is False:
            return data, False, 'close', 'error_not_logged'
        return data, True, 'valid', None

    @staticmethod
    def condense_stats_body(body):
        return {
            'job_id': body.get('job_id'),
            'job_group_id': body.get('job_group_id'),
        }

    @staticmethod
    def get(url, proxy=None, check=True):
        proxies = None
        if ProxyNormalizer.unknown_proxy_scheme(proxy) is not True:
            proxies = {ProxyNormalizer.get_proxy_scheme(proxy): proxy}
        for _ in range(SOPSRequest.RETRY_LIMIT):
            try:
                response = requests.get(url, timeout=SOPSRequest.TIMEOUT, proxies=proxies, headers={'api_key': SOPSRequest.API_KEY}) 
                if check and response.status_code == 401:
                    return None, 'invalid_api_key'
                if response.status_code == 200:
                    data = response.json()
                    return data, None
                else:
                    raise ScrapeOpsAPIResponseError
            except requests.exceptions.ConnectionError as e:
                error = e
                continue
            except ScrapeOpsAPIResponseError as e:
                error = e
                continue
            except Exception as e:
                error = e
                continue
        return None, str(error)


    @staticmethod
    def post(url, body=None, files=None, proxy=None):
        proxies = None
        if ProxyNormalizer.unknown_proxy_scheme(proxy) is not True:
            proxies = {ProxyNormalizer.get_proxy_scheme(proxy): proxy}
        for _ in range(SOPSRequest.RETRY_LIMIT):
            try:
                response = requests.post(url, json=body, timeout=SOPSRequest.TIMEOUT, files=files, proxies=proxies, headers={'api_key': SOPSRequest.API_KEY}) 
                if response.status_code == 401:
                    return None, 'invalid_api_key'
                if response.status_code == 200:
                    data = response.json()
                    return data, None
                else:
                    time.sleep(3)
                    raise ScrapeOpsAPIResponseError
            except requests.exceptions.ConnectionError as e:
                error = e
                continue
            except ScrapeOpsAPIResponseError as e:
                error = e
                continue
            except Exception as e:
                error = e
                continue
        return None, str(error)


    @staticmethod
    def post_file(url, body=None, files=None):
        for _ in range(SOPSRequest.RETRY_LIMIT):
            try:
                response = requests.post(url, data=body, timeout=SOPSRequest.TIMEOUT, files=files, headers={'api_key': SOPSRequest.API_KEY}) 
                if response.status_code == 401:
                    return None, 'invalid_api_key'
                if response.status_code == 200:
                    data = response.json()
                    return data, None
                else:
                    raise ScrapeOpsAPIResponseError
            except requests.exceptions.ConnectionError as e:
                error = e
                continue
            except ScrapeOpsAPIResponseError as e:
                error = e
                continue
            except Exception as e:
                error = e
                continue
        return None, str(error)
    
    


    






    
 



   


    







         
    




from twisted.internet import task

from scrapeops_scrapy.exceptions import ScrapeOpsMissingAPIKey
from scrapeops_scrapy.utils import utils
from scrapeops_scrapy.core.controllers import SDKControllers
from scrapeops_scrapy.stats.logger import StatsLogger
from scrapeops_scrapy.normalizer.request_response import RequestResponse


class ScrapeopsCore(SDKControllers, StatsLogger):
    """
        Where the core ScrapeOps Functionality Goes
    """

    def __init__(self):
        SDKControllers.__init__(self)
        StatsLogger.__init__(self)
        
    def start_sdk(self, spider=None, crawler=None):
        self.start_time = self.period_start_time = utils.current_time()
        self.initialize_SDK(spider, crawler=crawler)
        if self.check_api_key_present():
            self.send_setup_request()
            self.spider_open_stats()
            self.start_periodic_monitor()
        else:
            err = ScrapeOpsMissingAPIKey()
            self.deactivate_sdk(reason='no_api_key', error=err)
            raise err
    

    def close_sdk(self, spider=None, reason=None):
        if self.sdk_enabled():
            self.period_finish_time = utils.current_time()
            self.spider_close_stats(reason=reason, crawler=self.crawler)
            self.send_stats(periodic_stats=self._periodic_stats, overall_stats=self._overall_stats, stats_type='finished', reason=reason)
            self.close_periodic_monitor()


    def request_stats(self, request=None):
        if self.sdk_enabled():
            request.meta['sops_time'] = utils.current_time()
            request_response_object = RequestResponse(request=request)
            self.request_response_middleware.normalise_domain_proxy_data(request_response_object)
            self.add_missed_urls_callback(request)
            self.generate_request_stats(request_response_object, request=request)


    def response_stats(self, request=None, response=None):
        if self.sdk_enabled():
            request_response_object = RequestResponse(request=request, response=response)
            self.request_response_middleware.process(request_response_object, response) 
            self.generate_response_stats(request_response_object, request=request, response=response)


    def exception_stats(self, request=None, exception_class=None):
        if self.sdk_enabled():
            request_response_object = RequestResponse(request=request)
            self.request_response_middleware.normalise_domain_proxy_data(request_response_object)
            self.generate_exception_stats(request_response_object, request=request,  exception_class=exception_class)


    def item_stats(self, signal_type=None, item=None, response=None, spider=None):
        if self.sdk_enabled():
            request_response_object = RequestResponse(response=response)
            if response is not None:
                self.request_response_middleware.normalise_domain_proxy_data(request_response_object) 
            if signal_type == 'item_scraped':
                self.item_validation_middleware.validate(request_response_object, item)
            self.generate_item_stats(request_response_object, signal=signal_type, response=response)


    def add_missed_urls_callback(self, request):
        if request.errback is None:
            request.errback = self.failed_url_middleware.log_failure
            
    """
        PERIODIC MONITOR
    """
    def start_periodic_monitor(self):
        if self.sdk_enabled():
            self.loop = task.LoopingCall(self.periodic_monitor)
            self.periodic_loop = self.loop.start(1, now=False) # Start looping every 1 second (1.0).

    def periodic_monitor(self):
        period_time = utils.current_time()
        if self.get_runtime(time=period_time) % self.get_periodic_frequency() == 0:
            self.period_finish_time = period_time
            if self.sdk_enabled():
                self.aggregate_stats(crawler=self.crawler, middleware=self.scrapeops_middleware_enabled()) 
                self.send_stats(periodic_stats=self._periodic_stats, overall_stats=self._overall_stats, stats_type='periodic')
                self.reset_periodic_stats()
                self.period_start_time = utils.current_time()
                self.inc_value(self._overall_stats, 'periodic_runs') 
            elif self.periodic_monitor_active(): 
                self.close_periodic_monitor()

    def close_periodic_monitor(self):
        if self.periodic_monitor_active():
            self.loop.stop()

    def periodic_monitor_active(self):
        if self.loop is not None:
            if self.loop.running:
                return True
        return False
    
    def get_periodic_frequency(self):
        self.period_count = 0
        runtime = self.get_runtime()
        if self._period_freq_list is None:
            self.period_count = int(runtime//self._period_frequency)
            return self._period_frequency
        for index, row in enumerate(self._period_freq_list):
            if runtime > int(row.get('total_time')):
                if index == 0:
                    period_time = row.get('total_time') 
                else:
                    period_time = row.get('total_time') - self._period_freq_list[index - 1].get('total_time') 
                self.period_count += int(period_time/row.get('periodic_frequency'))
            if runtime <= int(row.get('total_time')):
                self._period_frequency = row.get('periodic_frequency')
                if index == 0:
                    diff = runtime
                else:
                    diff = runtime - int(self._period_freq_list[index - 1].get('total_time'))
                self.period_count += int(diff//self._period_frequency)
                return self._period_frequency 
        return self._period_frequency

    

    
    
 






    






    






    
    

    





    
        







from scrapeops_scrapy.utils import utils
from scrapeops_scrapy.core.setup import SDKSetup
from scrapeops_scrapy.core.api import SOPSRequest 

class SDKControllers(SDKSetup):

    SETUP_ATTEMPT_LIMIT = 3

    def __init__(self):
        SDKSetup.__init__(self)

    def send_setup_request(self):
        data, status = SOPSRequest().setup_request(body=self.setup_data())
        if status.valid:
            self.initialize_job_details(data)
        elif status.action == 'retry' and self._setup_attempts < SDKControllers.SETUP_ATTEMPT_LIMIT:
            self._setup_attempts += 1
            self._error_logger.log_error(reason='setup_failed', 
                                         error=status.error, 
                                         data={'setup_attempts': self._setup_attempts})
        elif status.action == 'retry' and self._setup_attempts >= SDKControllers.SETUP_ATTEMPT_LIMIT:
            self.deactivate_sdk(reason='exceeded_max_setup_attempts', 
                                error=status.error, 
                                data={'setup_attempts': self._setup_attempts}, 
                                request_type='setup')
        else:
            self.deactivate_sdk(reason=status.error, data=data, request_type='setup')


    def send_stats(self, periodic_stats=None, overall_stats=None, reason=None, stats_type=None):
        self._sdk_run_time = self._sdk_run_time + self._period_frequency 
        post_body = self.stats_data(periodic_stats=periodic_stats, overall_stats=overall_stats, stats_type=stats_type, reason=reason)

        if self.job_active() is False:
            self.send_setup_request()

        ## retest if job is inactive
        if self.job_active() is False:
            self.failed_periods += 1
            self.cache_failed_stats(post_body)
            self._error_logger.log_error(reason=f'sending_{stats_type}_stats_failure', 
                                        data={'failed_periods': self.failed_periods})

        if self.job_active():
            if stats_type == 'finished' and self.export_logs():
                log_body = self.log_data()
                with open(self.log_file, 'rb') as f:
                    data, status = SOPSRequest().stats_request(body=post_body, log_body=log_body, files={'file': f}) 
            else:
                data, status = SOPSRequest().stats_request(body=post_body) 
            
            if status.valid:
                self.update_sdk_settings(data) 
                self.reset_failed_stats()
            elif status.action == 'retry' and self.failed_periods < 3:
                self.failed_periods += 1
                self.cache_failed_stats(post_body)
                self._error_logger.log_error(reason=f'sending_{stats_type}_stats_failure', 
                                            error=status.error, 
                                            data={'failed_periods': self.failed_periods})
            else:
                self.deactivate_sdk(reason=f'sending_{stats_type}_stats_failure', error=status.error,
                                    data={'failed_periods': self.failed_periods}, request_type=stats_type)   


    def sdk_enabled(self):
        if self._sdk_active:
            if self.request_response_middleware is None:
                self.initialize_normalizer_middleware()
            return True
        return False


    def check_api_key_present(self):
        if self._scrapeops_api_key == None:
            self._sdk_active = False
            return False
        self._sdk_active = True
        return True

    def deactivate_sdk(self, reason=None, error=None, request_type=None, data=None):
        self._sdk_active = False
        self._error_logger.sdk_error_close(reason=reason, error=error, request_type=request_type)

    def job_active(self):
        if self.job_id is None and self._sdk_active:
            return False
        return True

    def cache_failed_stats(self, post_body):
        self.cached_failed_stats.append(post_body)
        self.failed_periods = len(self.cached_failed_stats)

    def reset_failed_stats(self):
        self.cached_failed_stats = []
        self.failed_periods = 0

    def get_runtime(self, time=None):
        if time is None:
            return utils.current_time() - self._scrapeops_job_start 
        return time - self._scrapeops_job_start 

    def scrapeops_middleware_enabled(self):
        if self._scrapeops_middleware is True:
            return True
        return False

    def export_logs(self):
        if self._scrapeops_export_scrapy_logs and self.log_file is not None:
            return True
        return False

    
    


import socket
import scrapy.settings.default_settings as default_settings


class BaseSDKModel(object):

    """
        SDK Model:
        The core data types used to control the SDK's operation. 
    """
    
    def __init__(self):
        ## User Data
        self._scrapeops_api_key = None

        ## SDK Data
        self._sdk_active = None
        self._scrapeops_endpoint = None
        self._scrapeops_middleware = None
        self._scrapeops_settings_exclusion_list = []
        self._scrapeops_export_scrapy_logs = False
        self._period_frequency = 60 
        self._period_freq_list = None
        self._sdk_run_time = 0
        self._setup_attempts = 0
        self._scrapeops_test_id = None
        self._error_logger = None
        self._scrapeops_sdk_version = None
        self._scrapeops_scrapy_version = None
        self._scrapeops_python_version = None
        self._scrapeops_system_version = None
        self._scrapeops_job_start = None
        
        ## Spider Data
        self.crawler = None
        self.spider = None
        self.spider_name = None
        self.spider_id= None
        self.spider_settings = None
        self.server_id= None
        self.project_id = None
        self.project_name = None
        self.bot_name = None
        self.retry_enabled = None
        self.retry_times = None
        self.log_file = None

        ## Overall Job Data
        self.job_args = None
        self.job_id = None
        self.job_group_id = None
        self.job_group_uuid = None
        self.job_group_name = None
        self.job_group_version = None
        self.start_time = None
        self.finish_time = None
        self.server_hostname = None
        self.server_ip = None
        self._proxy_apis  = {}
        self._generic_validators = {}
        self.multi_server = False
        self.failed_urls = []

        ## Period Data
        self.period_start_time = None
        self.period_finish_time = None
        self.period_run_time = 0
        self.period_concurrency = 0
        self.period_count = 0
        
        ## ScrapeOps Triggered Jobs
        self._scrapeops_server_id = None
        self.job_group_type = None

        ## Periodic Monitor
        self.loop = None
        self.periodic_loop = None

        ## Validation/Normalisation Data
        self.proxy_domains = []

        ## Failure 
        self.failed_periods = 0
        self.cached_failed_stats = []

        ## Middleware
        self.request_response_middleware = None
        self.item_validation_middleware = None
        self.failed_url_middleware = None

        self.allowed_response_codes = []


class SDKData(BaseSDKModel):

    def __init__(self):
        BaseSDKModel.__init__(self)


    def setup_data(self):
        return {
            'sops_api_key': self._scrapeops_api_key,
            'job_group_name': self.job_group_name,
            'job_group_version': self.job_group_version,
            'job_group_identifier': self.job_group_uuid,
            'job_group_type': self.job_group_type, 
            'job_settings': self.spider_settings,
            'job_args': self.job_args,
            'job_start_time': self.start_time,
            'sops_sdk': 'scrapy',
            'sops_scrapeops_version': self._scrapeops_sdk_version,
            'sops_scrapy_version': self._scrapeops_scrapy_version,
            'sops_python_version': self._scrapeops_python_version,
            'sops_system_version': self._scrapeops_system_version,
            'sops_middleware_enabled': self._scrapeops_middleware,
            'sops_test_id': self._scrapeops_test_id,
            'sops_server_id': self._scrapeops_server_id,
            'scrapeops_job_start': self._scrapeops_job_start,
            'spider_name': self.spider_name,
            'server_ip': self.server_ip,
            'server_hostname': self.server_hostname,
            'project_name': self.project_name, 
            'bot_name': self.bot_name, 
            'multi_server': self.multi_server,
            'retry_enabled': self.retry_enabled,
            'retry_times': self.retry_times,
        }
    

    def stats_data(self, periodic_stats=None, overall_stats=None, stats_type=None, reason=None):
        data = {
            'job_id': self.job_id,
            'job_group_id': self.job_group_id,
            'type': stats_type,
            'period_start_time': self.period_start_time,
            'period_finish_time': self.period_finish_time,
            'period_run_time': self._period_frequency, 
            'sdk_run_time': self._sdk_run_time,
            'periodic': periodic_stats,
            'overall': overall_stats,
            'cached_failed_stats': self.cached_failed_stats,
            'periodic_warnings': periodic_stats.get('log_count/WARNING', 0),
            'periodic_errors': periodic_stats.get('log_count/ERROR', 0),
            'periodic_criticals': periodic_stats.get('log_count/CRITICAL', 0),
            'multi_server': self.multi_server,
            'period_count': self.period_count,
            'data_coverage': self.item_validation_middleware.get_item_coverage_data(),
            'invalid_items_count': self.item_validation_middleware.get_num_invalid_items(),
            'field_coverage': self.item_validation_middleware.get_field_coverage(),
            'failed_urls_count': self.failed_url_middleware.get_url_count(),
            'failed_urls_enabled': self.failed_url_middleware.enabled(),
            'scrapy_stats': self.get_scrapy_stats(), 
        }

        if stats_type == 'finished':
            data['job_finish_time'] = self.period_finish_time
            data['job_status'] = stats_type
            data['job_finish_reason'] = reason
            data['failed_urls_list'] = self.failed_url_middleware.get_url_list()
            data['invalid_items_urls_list'] = self.item_validation_middleware.get_invalid_items_urls()
        return data
    

    def log_data(self):
        return {
            'job_group_id': self.job_group_id,
            'job_group_name': self.job_group_name,
            'job_group_identifier': self.job_group_uuid,
            'spider_name': self.spider_name,
            'sops_sdk': 'scrapy',
        }

    

    def logging_data(self):
        return {
            'sops_api_key': self._scrapeops_api_key,
            'job_id': self.job_id,
            'job_group_id': self.job_group_id,
            'job_group_identifier': self.job_group_uuid,
            'job_group_name': self.job_group_name,
            'spider_name': self.spider_name,
            'spider_id': self.spider_id,
            'server_id': self.server_id,
            'project_id': self.project_id,
            'project_name': self.project_name, 
            'bot_name': self.bot_name, 
            'server_ip': self.server_ip,
            'server_hostname': self.server_hostname,
            'sops_scrapeops_version': self._scrapeops_sdk_version,
            'sops_scrapy_version': self._scrapeops_scrapy_version,
            'sops_python_version': self._scrapeops_python_version,
            'sops_system_version': self._scrapeops_system_version,
            'sops_middleware_enabled': self._scrapeops_middleware,
            'sops_sdk': 'scrapy',
        }

    def check_spider_attributes(self, spider):
        if hasattr(spider, 'sops_test'):
            if spider.sops_test.test_active():
                self._scrapeops_test_id = spider.sops_test.generate_test_id()


    def get_settings(self, spider):
        default_scrapy_settings = default_settings.__dict__
        full_settings = spider.settings.copy_to_dict()
        self.spider_settings = {}
        for key, value in full_settings.items():
            if key not in default_scrapy_settings and self.include_setting(key):
                self.spider_settings[key] = value
            elif default_scrapy_settings.get(key) != value and self.include_setting(key):
                self.spider_settings[key] = value

    def include_setting(self, key):
        exclusion_terms = ['API_KEY', 'APIKEY', 'SECRET_KEY', 'SECRETKEY', 'PASSWORD', 'CONNECTION_STRING']
        if key in self._scrapeops_settings_exclusion_list:
            return False
        for term in exclusion_terms:
            if term in key.upper(): return False 
        return True

    
    def get_job_name(self):
        ## check args
        for arg in self.job_args.get('args'):
            if 'SCRAPEOPS_JOB_NAME' in arg:
                return arg.split('=')[1]

        ## check spider defined
        if hasattr(self.spider, 'sops_job_name'):
            return self.spider.sops_job_name
        if hasattr(self.spider, 'name'):
            return self.spider.name
        return 'no_spider_name'


    def get_job_version(self):
        ## check args
        for arg in self.job_args.get('args'):
            if 'SCRAPEOPS_JOB_VERSION' in arg:
                return arg.split('=')[1]

        ## check spider defined
        if hasattr(self.spider, 'sops_job_version'):
            return self.spider.sops_job_version
        return 'default'

    
    def get_server_id(self, crawler):
        for arg in self.job_args.get('args'):
            if 'SCRAPEOPS_SERVER_ID' in arg:
                return arg.split('=')[1]
        if crawler.settings.get('SCRAPEOPS_SERVER_ID') is not None:
            return crawler.settings.get('SCRAPEOPS_SERVER_ID')
        return '-1'

    
    def check_scrapeops_triggered_job(self, crawler):
        self._scrapeops_server_id = self.get_server_id(crawler)
        if isinstance(self._scrapeops_server_id, str) is False: self._scrapeops_server_id = str(self._scrapeops_server_id)
        if self._scrapeops_server_id != '-1':
            self.job_group_type = 'scrapeops_triggered'
        else:
            self.job_group_type = 'user_triggered'

    def get_server_details(self):
        try:
            self.server_hostname = socket.gethostname()
            self.server_ip = socket.gethostbyname(self.server_hostname)
        except Exception:
            self.server_hostname = 'unknown'
            self.server_ip = 'unknown'

    
    def get_uuid(self):
        for arg in self.job_args.get('args'):
            if 'SCRAPEOPS_JOB_GROUP_IDENTIFIER' in arg:
                return arg.split('=')[1]
        if hasattr(self.spider, 'sops_job_group_identifier'):
            return self.spider.sops_job_group_identifier
        self.multi_server = False
        return ''


    def get_export_logs(self, crawler):
        for arg in self.job_args.get('args'):
            if 'SCRAPEOPS_EXPORT_SCRAPY_LOGS' in arg:
                try: 
                    if arg.split('=')[1] == 'True':
                        return True
                except Exception:
                    pass
        if crawler.settings.get('SCRAPEOPS_EXPORT_SCRAPY_LOGS') is not None:
            return True
        return False

    def get_scrapy_stats(self):
        scrapy_stats = self.crawler.stats.get_stats()
        return {k:str(v) for (k,v) in scrapy_stats.items()}




from scrapeops_scrapy.core.api import SOPSRequest 
from scrapeops_scrapy.validators.response_validator import ResponseValidator
from scrapeops_scrapy.normalizer.proxies import ProxyNormalizer
from scrapeops_scrapy.normalizer.proxy_port_normalizer import ProxyPortStringNormalizer


class RequestResponseMiddleware(object):

    PROXY_DOMAIN_NORMALIZATION = True
    RESPONSE_VALIDATION = True
    PROXY_ALERTS = False
    FAILED_URL_LOGGER_ENABLED = True
    LOG_MISSED_URLS = False

    def __init__(self, job_group_id, proxy_apis, generic_validators, error_logger, allowed_response_codes):
        self.job_group_id = job_group_id
        self._proxy_apis = proxy_apis
        self._data_coverage_validation = False
        self._domains = {}
        self._proxies = {}
        self._proxy_port_setups = {}
        self._generic_validators = generic_validators
        self._allowed_response_codes = allowed_response_codes
        self._error_logger = error_logger
        self._error_count = 0
        self._error_alerts_sent = {}
        self._missed_urls = {}


    def process(self, request_response_object, response):
        self.normalise_domain_proxy_data(request_response_object)
        self.check_proxy_responses(request_response_object, response)
        self.validate_response_data(request_response_object, response)


    def normalise_domain_proxy_data(self, request_response_object):
        if RequestResponseMiddleware.PROXY_DOMAIN_NORMALIZATION:

            proxy_api = self.normalise_proxy_api(request_response_object)
            if proxy_api is False:
                self.normalise_proxy_port(request_response_object)
            self.normalise_domain_data(request_response_object)

        if RequestResponseMiddleware.PROXY_DOMAIN_NORMALIZATION is False:
            request_response_object.fallback_domain_proxy_details(reason='disabled')
    

    def normalise_proxy_api(self, request_response_object):
        try:
            proxy_api, update = request_response_object.check_proxy_api(self._proxy_apis)
            if proxy_api and update:
                data, status = SOPSRequest().proxy_api_normalisation_request(request_response_object)
                if status.valid:
                    self._proxy_apis[request_response_object.get_proxy_api_name()] = data.get('proxy_parsing_data')
                    request_response_object.update_proxy_api(data.get('proxy_parsing_data'))
                else:
                    if self._proxy_apis.get(request_response_object.get_proxy_api_name()) is None:
                            self._proxy_apis[request_response_object.get_proxy_api_name()] = {}
                    self._proxy_apis[request_response_object.get_proxy_api_name()]['proxy_setup'] = {}
                    self._error_logger.log_error(reason='get_proxy_api_details_failed', 
                                            error=status.error, 
                                            data={'proxy_api': request_response_object.get_proxy_api_name()})
                    request_response_object.fallback_proxy_details(proxy_type='proxy_api', proxy_apis=self._proxy_apis)
            
        except Exception:
            request_response_object.fallback_proxy_details(proxy_type='proxy_api', proxy_apis=self._proxy_apis)
        
        return proxy_api or False


    def normalise_proxy_port(self, request_response_object):
        try:
            if request_response_object.active_proxy_port():
                named_proxy, update = request_response_object.check_proxy_port_type(self._proxies)
                if named_proxy and update:
                    data, status = SOPSRequest().proxy_port_normalisation_request(request_response_object)
                    if status.valid:
                        ProxyNormalizer.update_proxy_details(self._proxies, request_response_object, data, valid=True)
                        ProxyPortStringNormalizer.proxy_port_test(self._proxies, request_response_object, data, valid=True)
                    else:
                        ProxyNormalizer.update_proxy_details(self._proxies, request_response_object, data, valid=False)
                        self._error_logger.log_error(reason='get_proxy_port_details_failed', 
                                                error=status.error, 
                                                data={'proxy_port': request_response_object.get_raw_proxy()})
                
            ## Using No Proxy
            if request_response_object.active_proxy() is False:
                request_response_object.update_no_proxy()

        except Exception:
            request_response_object.fallback_proxy_details(proxy_type='proxy_port')


    def normalise_domain_data(self, request_response_object):
        try:
            ## Normalise domain/page type data
            unknown = request_response_object.check_domain(self._domains)
            if unknown:
                data, status = SOPSRequest().domain_normalisation_request(request_response_object)
                if status.valid:
                    self._domains[request_response_object.get_domain()] = data.get('domain_parsing_data')
                    request_response_object.update_page_type(data.get('domain_parsing_data'))
                else:
                    if self._domains.get(request_response_object.get_domain()) is None:
                        self._domains[request_response_object.get_domain()] = {}
                    self._domains[request_response_object.get_domain()]['url_contains_page_types'] = {}
                    self._domains[request_response_object.get_domain()]['query_param_page_types'] = {}
                    self._domains[request_response_object.get_domain()]['validation_details'] = [] 
                    self._error_logger.log_error(reason='get_domain_details_failed', 
                                            error=status.error, 
                                            data={'real_url': request_response_object.get_real_url()})
                    request_response_object.fallback_domain_data()
        
        except Exception:
            request_response_object.fallback_domain_data()


    def check_proxy_responses(self, request_response_object, response):
        if RequestResponseMiddleware.PROXY_ALERTS:
            if request_response_object.active_proxy_api():
                proxy_details = self._proxy_apis.get(request_response_object.get_proxy_api_name())
                if proxy_details is not None:
                    self.check_proxy_error_codes(request_response_object, proxy_details, response)
            
            if request_response_object.active_named_proxy():
                proxy_details = self._proxies.get(request_response_object.get_proxy_port_name())
                if proxy_details is not None:
                    self.check_proxy_error_codes(request_response_object, proxy_details, response)


    def check_proxy_error_codes(self, request_response_object, proxy_details, response):
        error_codes = proxy_details.get('error_codes')
        if error_codes is not None:
            status_code = str(response.status)
            error_response = error_codes.get(status_code)
            if error_response is not None:
                if error_response.get('action') == 'alert' and self.should_alert(error_response, status_code):
                    _, status = SOPSRequest().proxy_alert_request(request_response_object, self.job_group_id, error_response, self._error_alerts_sent.get(status_code))
                    if status.valid:
                        self._error_alerts_sent[status_code] += 1
                elif error_response.get('action') == 'monitor':
                    self._error_count += 1
                    if self._error_count  > error_response.get('error_limit', 0) and self.should_alert(error_response, status_code): 
                        _, status = SOPSRequest().proxy_alert_request(request_response_object, self.job_group_id, error_response, self._error_alerts_sent.get(status_code))
                    if status.valid:
                        self._error_alerts_sent[status_code] += 1


    def should_alert(self, error_response, status_code):
        if self._error_alerts_sent.get(status_code) is None:
            self._error_alerts_sent[status_code] = 0
            return True
        if self._error_alerts_sent.get(status_code) is not None:
            if self._error_alerts_sent[status_code] < error_response.get('alert_limit'):
                return True
        return False


    def validate_response_data(self, request_response_object, response=None):
        if RequestResponseMiddleware.RESPONSE_VALIDATION and response is not None:
            if response.status == 200:
                domain_tests = ResponseValidator.get_domain_tests(request_response_object, self._domains)
                ResponseValidator.validate(request_response_object, response, domain_tests=domain_tests, generic_tests=self._generic_validators)

            if response.status != 200 and ResponseValidator.failed_scan(request_response_object, self._domains):
                ResponseValidator.validate(request_response_object, response, generic_tests=self._generic_validators)


    def failed_url(self, request_response_object, response=None):
        if RequestResponseMiddleware.FAILED_URL_LOGGER_ENABLED:
            if (response.status < 200 and response.status > 300) and (response.status not in self._allowed_response_codes):
                if self._missed_urls.get('count') is None:
                    self._missed_urls['count'] = 0
                self._missed_urls['count'] += 1

                if RequestResponseMiddleware.LOG_MISSED_URLS:
                    if self._missed_urls.get(response.status) is None:
                        self._missed_urls[response.status] = []
                    self._missed_urls[response.status].append(request_response_object.get_real_url())








        
        


        

        






from tld import get_tld
from urllib.parse import urlparse, parse_qs

class DomainNormalizer(object):

    def __init__(self):
        pass
    
    @staticmethod
    def get_domain(url):
        #if 'http://' not in url or 'http://' not in url or 'socks5://' not in url
        try:
            if DomainNormalizer.if_localhost(url):
                return 'localhost'
            res = get_tld(url, as_object=True)
            return res.fld
        except Exception:
            return 'unknown'

    @staticmethod
    def get_full_domain(url):
        try:
            if DomainNormalizer.if_localhost(url):
                return 'localhost'
            res = get_tld(url, as_object=True)
            if res.subdomain != '':
                return res.subdomain + '.' + res.fld
            return res.fld
        except Exception:
            return 'unknown'


    @staticmethod
    def if_localhost(url):
        if 'http://localhost:' in url or 'http://127.0.0.1:' in url:
                return True
        return False


    @staticmethod
    def parse_url(url):
        parsed_url = urlparse(url)
        query_params = parse_qs(parsed_url.query)
        query_dict = {}
        for key, value in query_params.items():
            query_dict[key] = value[0]
        return query_dict


    @staticmethod
    def get_url_proxy_api(url=None, proxy_settings=None):
        url_identifier = proxy_settings.get('url_identifier')
        query_params = DomainNormalizer.parse_url(url)
        url = query_params.get(url_identifier)
        return url

    
    @staticmethod
    def get_page_type(url, domain_data):
        if domain_data.get('url_classification'):
            url_classifiers = domain_data.get('url_contains_page_types', {})
            for k, v in url_classifiers.items():
                if k in url:
                    return v
            query_param_page_types = domain_data.get('query_param_page_types', {})
            query_params = DomainNormalizer.parse_url(url)
            for k, v in query_params.items():
                key_mapping = query_param_page_types.get(k, None)
                if key_mapping is not None:
                    return v
        return 'none'    
    
 



   


    







         
    
